var b = require('../core')

var NSMutableArray = b.objc_getClass('NSMutableArray')
  , NSString = b.objc_getClass('NSString')
  , NSAutoreleasePool = b.objc_getClass('NSAutoreleasePool')

var alloc = b.sel_registerName('alloc')
  , init = b.sel_registerName('init')
  , describe = b.sel_registerName('description')
  , UTF8String = b.sel_registerName('UTF8String')
  , addObject = b.sel_registerName('addObject:')
  , sortUsingFunction = b.sel_registerName('sortUsingFunction:context:')
  , enumerateObjectsUsingBlock = b.sel_registerName('enumerateObjectsUsingBlock:')
  , UTF8StringMethod = b.class_getInstanceMethod(NSString, UTF8String)
  , allocMethod = b.class_getClassMethod(NSMutableArray, alloc)
  , addObjectMethod = b.class_getInstanceMethod(NSMutableArray, addObject)
  , sufMethod = b.class_getInstanceMethod(NSMutableArray, sortUsingFunction)
  , enumMethod = b.class_getInstanceMethod(NSMutableArray, enumerateObjectsUsingBlock)

var msgSend = b.get_objc_msgSend(getTypes(allocMethod))
  , msgSend2 = b.get_objc_msgSend(getTypes(UTF8StringMethod))
  , msgSend3 = b.get_objc_msgSend(getTypes(addObjectMethod))
  , msgSend4 = b.get_objc_msgSend(getTypes(sufMethod))
  , msgSend5 = b.get_objc_msgSend(getTypes(enumMethod))
  , pool = msgSend(msgSend(NSAutoreleasePool, alloc), init)
  , instance = msgSend(msgSend(NSMutableArray, alloc), init)

msgSend3(instance, addObject, NSString);
msgSend3(instance, addObject, NSMutableArray);


// toString() before sort
console.error(msgSend2(msgSend(instance, describe), UTF8String));

// we can sort, using a JavaScript function to do the sorting logic!!!
// In this simple example we sort based on the length of the class name
var callback = new b.Callback([ 'int32', [ 'pointer', 'pointer', 'pointer' ]], function (obj1, obj2, context) {
  var n1 = b.class_getName(obj1)
    , n2 = b.class_getName(obj2)
  return n1.length < n2.length ? -1
       : n2.legnth > n1.length ? 1
       :  0;
});

msgSend4(instance, sortUsingFunction, callback.getPointer(), instance);


// toString() after sort
console.error(msgSend2(msgSend(instance, describe), UTF8String));

// We have to simulate what the llvm compiler does when it encounters a Block
// literal expression:
var __block_literal_1 = b.Struct([
    ['pointer', 'isa']
  , ['int32', 'flags']
  , ['int32', 'reserved']
  , ['pointer', 'invoke']
  , ['pointer', 'descriptor']
])

var __block_descriptor_1 = b.Struct([
    ['ulonglong', 'reserved']
  , ['ulonglong', 'Block_size']
])

// Enumerate using a block.
var blockFunc = new b.Callback([ 'void', [ 'pointer', 'pointer', 'uint32', 'pointer' ]], function (block, obj, index, stopPtr) {
  console.error('inside block!');
  console.error("Enumerate: %d!", index);
});

var invokePtr = blockFunc.getPointer()
  , bl = new __block_literal_1
  , bd = new __block_descriptor_1

// static
bd.reserved = 0
bd.Block_size = __block_literal_1.__structInfo__.size
//console.log(bd);

bl.isa = b.dlopen().get('_NSConcreteStackBlock')
bl.flags = (1<<29)
bl.reserved = 0
bl.invoke = invokePtr
bl.descriptor = bd.__pointer

//console.log(bl);
console.error(msgSend2(msgSend(bl.__pointer, describe), UTF8String));

msgSend5(instance, enumerateObjectsUsingBlock, bl.__pointer)




function getTypes (method) {
  if (method.isNull()) throw new Error('bad pointer!');
  var args = []
    , types = []
    , numArgs = b.method_getNumberOfArguments(method)
    , rtnTypePtr = b.method_copyReturnType(method)
    , rtnType = rtnTypePtr.getCString()
  b.free(rtnTypePtr);
  types.push(rtnType);
  types.push(args);
  for (var i=0; i<numArgs; i++) {
    var argPtr = b.method_copyArgumentType(method, i)
    args.push(argPtr.getCString());
    b.free(argPtr);
  }
  return types;
}
